from pprint import pprint
from .common import connect, Context

def list(context: Context, args: list):
    if len(args) == 0:
        print("ERROR: no library ID was provided")
        return
    gi = connect(context)
    folders = gi.libraries.get_folders(args[0])
    pprint(folders)


def create(context: Context, args: list):
    if len(args) < 2:
        print("ERROR: Invalid parameters.  Required the library ID, folder name and folder description (optional")
        return
    library_id = args[0]
    folder_name = args[1]
    folder_description = "No description available"
    if len(args) > 2:
        folder_description = args[2]
    gi = connect(context)
    result = gi.libraries.create_folder(library_id, folder_name, folder_description)
    pprint(result)


def delete(context: Context, args: list):
    print("This functionality has not been implemented yet.")





import os
import sys
import subprocess
from ruamel.yaml import YAML
import json
import bioblend.galaxy
from bioblend.galaxy import dataset_collections

import lib

PROFILE_SEARCH_PATH = ['~/.abm/profile.yml', '.abm-profile.yml']

datasets = {
    "dna": [
        "ftp://ftp.sra.ebi.ac.uk/vol1/fastq/ERR013/ERR013101/ERR013101_1.fastq.gz",
        "ftp://ftp.sra.ebi.ac.uk/vol1/fastq/ERR015/ERR015526/ERR015526_2.fastq.gz",
        "ftp://ftp.sra.ebi.ac.uk/vol1/fastq/SRR592/SRR592109/SRR592109_2.fastq.gz",
        "ftp://ftp.sra.ebi.ac.uk/vol1/fastq/SRR233/SRR233167/SRR233167_2.fastq.gz",
        "ftp://ftp.sra.ebi.ac.uk/vol1/fastq/ERR047/ERR047678/ERR047678_1.fastq.gz",
        "ftp://ftp-trace.ncbi.nlm.nih.gov/ReferenceSamples/giab/data/AshkenazimTrio/HG003_NA24149_father/NIST_Illumina_2x250bps/reads/D2_S1_L002_R1_002.fastq.gz"
    ],
    "rna": []
}

class Context:
    def __init__(self, *args):
        if len(args) == 1:
            arg = args[0]
            if type(arg) == str:
                self.GALAXY_SERVER, self.API_KEY, self.KUBECONFIG = parse_profile(arg)
            elif type(arg) == dict:
                self.GALAXY_SERVER = arg['GALAXY_SERVER']
                self.API_KEY = arg['API_KEY']
                self.KUBECONFIG = arg['KUBECONFIG']
            else:
                raise Exception(f'Invalid arg for Context: {type(arg)}')
        elif len(args) == 3:
            self.GALAXY_SERVER = args[0]
            self.API_KEY = args[1]
            self.KUBECONFIG = args[2]
        else:
            raise Exception(f'Invalid args for Context. Expected one or three, found {len(args)}')



def print_json(obj, indent=2):
    print(json.dumps(obj, indent=indent))


def print_yaml(obj):
    get_yaml_parser().dump(obj, sys.stdout)


def connect(context:Context):
    """
    Create a connection to the Galaxy instance

    :return: a GalaxyInstance object
    """
    if context.GALAXY_SERVER is None:
        print('ERROR: The Galaxy server URL has not been set.  Please check your')
        print('       configuration in ~/.abm/profile.yml and try again.')
        sys.exit(1)
    if context.API_KEY is None:
        print('ERROR: The Galaxy API key has not been set.  Please check your')
        print('       configuration in ~/.abm/profile.yml and try again.')
        sys.exit(1)
    gi = bioblend.galaxy.GalaxyInstance(url=context.GALAXY_SERVER, key=context.API_KEY)
    gi.max_get_attempts = 3
    gi.get_retry_delay = 1
    return gi



def _set_active_profile(profile_name: str):
    # print(f"Parsing profile for {profile_name}")
    lib.GALAXY_SERVER, lib.API_KEY, lib.KUBECONFIG = parse_profile(profile_name)
    return lib.GALAXY_SERVER != None


def get_context(profile_name: str):
    return Context(profile_name)


def get_yaml_parser():
    if lib.parser is None:
        lib.parser = YAML()
    return lib.parser


def load_profiles():
    '''
    Load the profile configuration file.

    :return: a dictionary containing the YAML content of the configuration.
    '''
    yaml = get_yaml_parser()
    profiles = {}
    for profile_path in PROFILE_SEARCH_PATH:
        profile_path = os.path.expanduser(profile_path)
        if os.path.exists(profile_path):
            with open(profile_path, 'r') as f:
                # print(f'Loading profile from {profile_path}')
                profiles = yaml.load(f)
            break
    return profiles


def save_profiles(profiles: dict):
    yaml = get_yaml_parser()
    for profile_path in PROFILE_SEARCH_PATH:
        path = os.path.expanduser(profile_path)
        if os.path.exists(path):
            with open(path, 'w') as f:
                yaml.dump(profiles, f)
            print(f"Saved profiles to {path}")
            return


def parse_profile(profile_name: str):
    '''
    Parse the profile containing Galaxy URLs and API keys.

    :param profile_name: path to the profile to parse
    :return: a tuple containing the Galaxy URL, API key, and path to the kubeconfig
    '''
    profiles = load_profiles()
    if profiles is None:
        print(f'ERROR: Could not locate an abm profile file in {PROFILE_SEARCH_PATH}')
        return None, None, None
    if profile_name not in profiles:
        print(f'ERROR: {profile_name} is not the name of a valid profile.')
        keys = list(profiles.keys())
        if len(keys) > 0:
            quoted_keys = ', '.join([f"'{k}'" for k in keys[0:-2]]) + f", and '{keys[-1]}'"
            print(f'The defined profile names are: {quoted_keys}')
        return None, None, None
    profile = profiles[profile_name]
    if 'kube' in profile:
        return (profile['url'], profile['key'], os.path.expanduser(profile['kube']))
    return (profile['url'], profile['key'], None)


def run(command, env:dict= None):
    if env is None:
        env = os.environ
    # if env is not None:
    #     for name,value in env.items():
    #         os.environ[name] = value
    # if lib.KUBECONFIG is not None:
    #     os.environ['KUBECONFIG'] = lib.KUBECONFIG
    # local_env = os.environ.copy()
    # local_env.update(env)
    result = subprocess.run(command.split(), capture_output=True, env=env)
    if result.returncode != 0:
        raise RuntimeError(result.stderr.decode('utf-8').strip())
    return result.stdout.decode('utf-8').strip()


def get_env(context: Context):
    copy = os.environ.copy()
    for key,value in context.__dict__.items():
        if value is not None:
            copy[key] = value
    return copy


def find_executable(name):
    return run(f"which {name}")


# This is all the job metrics returned by Galaxy
#    "cpuacct.usage",
#     "end_epoch",
#     "galaxy_memory_mb",
#     "galaxy_slots",
#     "memory.failcnt",
#     "memory.limit_in_bytes",
#     "memory.max_usage_in_bytes",
#     "memory.memsw.limit_in_bytes",
#     "memory.memsw.max_usage_in_bytes",
#     "memory.oom_control.oom_kill_disable",
#     "memory.oom_control.under_oom",
#     "memory.soft_limit_in_bytes",
#     "memtotal",
#     "processor_count",
#     "runtime_seconds",
#     "start_epoch",
#     "swaptotal",
#     "uname"

def summarize_metrics(gi, jobs: list):
    header= [
    "id",
    "history_id",
    "history_name",
    "state",
    "tool_id",
    "invocation_id",
    "workflow_id",
    "cpuacct.usage",
    # "end_epoch",
    "galaxy_memory_mb",
    "galaxy_slots",
    # "memory.failcnt",
    "memory.limit_in_bytes",
    "memory.max_usage_in_bytes",
    # "memory.memsw.limit_in_bytes",
    # "memory.memsw.max_usage_in_bytes",
    # "memory.oom_control.oom_kill_disable",
    # "memory.oom_control.under_oom",
    "memory.soft_limit_in_bytes",
    "memtotal",
    "processor_count",
    "runtime_seconds",
    # "start_epoch",
    # "swaptotal",
    # "uname"
    ]

    print(','.join(header))
    for job in jobs:
        job_metrics = gi.jobs.get_metrics(job['id'])
        row = []
        metrics = metrics_to_dict(job_metrics, header)
        metrics['id'] = job['id']
        metrics['history_id'] = job['history_id']
        metrics['history_name'] = job['history_name']
        metrics['state'] = job['state']
        metrics['tool_id'] = job['tool_id']
        metrics['invocation_id'] = job['invocation_id']
        for key in header:
            if key in metrics:
                row.append(metrics[key])
            else:
                row.append('')
        print(','.join(row), end='\n')


def metrics_to_dict(metrics: list, accept:list):
    result = dict()
    for m in metrics:
        key = m['name']
        if key in accept:
            result[key] = m['raw_value']
    return result


def get_keys(d: dict):
    result = []
    for key in d.keys():
        result.append(key)
    result.sort()
    return result


def find_history(gi, name_or_id):
    history = gi.histories.get_histories(name=name_or_id)
    if history is None:
        return name_or_id
    if len(history) == 0:
        return None
    return history[0]['id']


def _get_dataset_data(gi, name_or_id):
    def make_result(data):
        return {
            'id': data['id'],
            'size': data['file_size'],
            'history': data['history_id']
        }

    try:
        ds = gi.datasets.show_dataset(name_or_id)
        return make_result(ds)
    except Exception as e:
        pass

    try:
        datasets = gi.datasets.get_datasets(name=name_or_id)  # , deleted=True, purged=True)
        for ds in datasets:
            # print_json(ds)
            state = True
            if 'state' in ds:
                state = ds['state'] == 'ok'
            if state and not ds['deleted'] and ds['visible']:
                # The dict returned by get_datasets does not include the input
                # file sizes so we need to make another call to show_datasets
                return make_result(gi.datasets.show_dataset(ds['id']))
    except Exception as e:
        pass

    return None


def _make_dataset_element(name, value):
    # print(f"Making dataset element for {name} = {value}({type(value)})")
    return dataset_collections.HistoryDatasetElement(name=name, id=value)






from lib.common import get_env, Context


def test_get_env():
    context = Context('server', 'key', 'kubeconfig')
    assert context.GALAXY_SERVER == 'server'
    assert context.API_KEY == 'key'
    assert context.KUBECONFIG == 'kubeconfig'
    env = get_env(context)
    assert env is not None




from common import Context, connect, summarize_metrics, print_json

def doList(context: Context, args: list):
    wid = None
    hid = None
    while len(args) > 0:
        arg = args.pop(0)
        if args in ['-h', '--history']:
            hid = args.pop(0)
        elif arg in ['-w', '--workflow']:
            wid = args.pop(0)
        else:
            print(f'Invalid parameter: {arg}')
            return
    gi = connect(context)
    invocations = gi.invocations.get_invocations(workflow_id=wid, history_id=hid)
    print('ID\tState\tWorkflow\tHistory')
    for invocation in invocations:
        id = invocation['id']
        state = invocation['state']
        workflow = invocation['workflow_id']
        history = invocation['history_id']
        print(f'{id}\t{state}\t{workflow}\t{history}')


def summarize(context: Context, args: list):
    if len(args) == 0:
        print("ERROR: Provide one or more invocation ID values.")
        return
    gi = connect(context)
    id = args[0]
    all_jobs = []
    jobs = gi.jobs.get_jobs(invocation_id=id)
    for job in jobs:
        job['invocation_id'] = id
        job['workflow_id'] = ''
        all_jobs.append(job)
    summarize_metrics(gi, all_jobs)



import os
import threading

import yaml
import json
import helm
import benchmark
import logging
import traceback
from common import load_profiles, Context
from time import perf_counter
from datetime import timedelta

INVOCATIONS_DIR = "invocations"
METRICS_DIR = "metrics"

log = logging.getLogger('abm')

def run(context: Context, args: list):
    """
    Runs a single benchmark defined by *args[0]*

    :param args: a list that contains a single element, the path to a benchmark
      configuration file.

    :return: True if the benchmarks completed sucessfully. False otherwise.
    """

    if len(args) == 0:
        print("ERROR: No benchmarking configuration provided.")
        return False

    benchmark_path = args[0]
    if not os.path.exists(benchmark_path):
        print(f"ERROR: Benchmarking configuration not found {benchmark_path}")
        return False

    with open(benchmark_path, 'r') as f:
        config = yaml.safe_load(f)

    profiles = load_profiles()
    # latch = CountdownLatch(len(config['cloud']))
    threads = []
    start = perf_counter()
    for cloud in config['cloud']:
        if cloud not in profiles:
            print(f"WARNING: No profile found for {cloud}")
            continue
        t = threading.Thread(target=run_on_cloud, args=(cloud, config))
        threads.append(t)
        print(f"Starting thread for {cloud}")
        t.start()
    print('Waiting for threads')
    for t in threads:
        t.join()

    end = perf_counter()
    print('All threads have terminated.')
    print(f"Execution time {timedelta(seconds=end - start)}")


def run_on_cloud(cloud: str, config: dict):
    print("------------------------")
    print(f"Benchmarking: {cloud}")
    context = Context(cloud)
    namespace = 'galaxy'
    chart = 'anvil/galaxykubeman'
    if 'galaxy' in config:
        namespace = config['galaxy']['namespace']
        chart = config['galaxy']['chart']
    if 'job_configs' in config and len(config['job_configs']) > 0:
        for conf in config['job_configs']:
            if not helm.update(context, [f"rules/{conf}.yml", namespace, chart]):
                log.warning(f"job configuration not found: rules/{conf}.yml")
                continue
            for n in range(config['runs']):
                history_name_prefix = f"{n+1} {cloud} {conf}"
                for workflow_conf in config['benchmark_confs']:
                    benchmark.run(context, workflow_conf, history_name_prefix, config['name'])
    else:
        for n in range(config['runs']):
            history_name_prefix = f"{n+1} {cloud}"
            for workflow_conf in config['benchmark_confs']:
                benchmark.run(context, workflow_conf, history_name_prefix, config['name'])


def test(context: Context, args: list):
    print(context.GALAXY_SERVER)
    if os.path.exists(args[0]):
        with open(args[0]) as f:
            data = yaml.safe_load(f)
        print(data)


def parse_toolid(id:str) -> str:
    parts = id.split('/')
    return f"{parts[-2]},{parts[-1]}"


def summarize(context: Context, args: list):
    """
    Parses all the files in the specified directory and prints metrics
    as CSV to stdout

    :param args[0]: The path to the directory containing metrics filees
    :return: None
    """
    separator = None
    input_dirs = []
    make_row = make_table_row
    header_row = "Run,Cloud,Job Conf,Workflow,History,Inputs,Tool,Tool Version,State,Slots,Memory,Runtime (Sec),CPU,Memory Limit (Bytes),Memory Max usage (Bytes)"
    for arg in args:
        if arg in ['-t', '--tsv']:
            if separator is not None:
                print('ERROR: The output format is specified more than once')
                return
            print('tsv')
            separator = '\t'
        elif arg in ['-c', '--csv']:
            if separator is not None:
                print('ERROR: The output format is specified more than once')
                return
            separator = ','
            print('csv')
        elif arg in ['-m', '--model']:
            if separator is not None:
                print('ERROR: The output format is specified more than once')
                return
            print('making a model')
            separator = ','
            make_row = make_model_row
            header_row = "job_id,tool_id,tool_version,state,memory.max_usage_in_bytes,cpuacct.usage,process_count,galaxy_slots,runtime_seconds,ref_data_size,input_data_size_1,input_data_size_2"
        else:
            print(f"Input dir {arg}")
            input_dirs.append(arg)

    if len(input_dirs) == 0:
        input_dirs.append('metrics')

    if separator is None:
        separator = ','

    print(header_row)
    for input_dir in input_dirs:
        for file in os.listdir(input_dir):
            input_path = os.path.join(input_dir, file)
            if not os.path.isfile(input_path) or not input_path.endswith('.json'):
                continue
            try:
                with open(input_path, 'r') as f:
                    data = json.load(f)
                if data['metrics']['tool_id'] == 'upload1':
                    #print('Ignoring upload tool')
                    continue
                row = make_row(data)
                print(separator.join([ str(x) for x in row]))
            except Exception as e:
                # Silently fail to allow the remainder of the table to be generated.
                print(f"Unable to process {input_path}")
                print(e)
                traceback.print_exc( )
                #pass


accept_metrics = ['galaxy_slots', 'galaxy_memory_mb', 'runtime_seconds', 'cpuacct.usage','memory.limit_in_bytes', 'memory.max_usage_in_bytes']  #,'memory.soft_limit_in_bytes']

def make_table_row(data: dict):
    row = [ str(data[key]) for key in ['run', 'cloud', 'job_conf', 'workflow_id', 'history_id', 'inputs']]
    row.append(parse_toolid(data['metrics']['tool_id']))
    row.append(data['metrics']['state'])
    for e in _get_metrics(data['metrics']['job_metrics']):
        row.append(e)
    return row


def make_model_row(data: dict):
    metrics = data['metrics']
    row = []
    row.append(metrics['id'])
    tool_id = metrics['tool_id']
    row.append(tool_id)
    row.append(tool_id.split('/')[-1])
    row.append(metrics['state'])
    job_metrics = parse_job_metrics(metrics['job_metrics'])
    row.append(job_metrics.get('memory.max_usage_in_bytes',0))
    row.append(job_metrics.get('cpuacct.usage', 0))
    row.append(job_metrics.get('processor_count', 0))
    row.append(job_metrics.get('galaxy_slots', 0))
    row.append(job_metrics.get('runtime_seconds', 0))
    if len(data['ref_data_size']) == 0:
        row.append('0')
    else:
        row.append(data['ref_data_size'][0])
    for size in data['input_data_size']:
        row.append(size)
    return row

def _get_metrics(metrics: list):
    row = [''] * len(accept_metrics)
    for job_metrics in metrics:
        if job_metrics['name'] in accept_metrics:
            index = accept_metrics.index(job_metrics['name'])
            try:
                row[index] = job_metrics['raw_value']
            except:
                pass
    return row

def add_metrics_to_row(metrics_list: list, row: list):
    for job_metrics in metrics_list:
        if job_metrics['name'] in accept_metrics:
            index = accept_metrics.index(job_metrics['name'])
            try:
                row[index + 8] = job_metrics['raw_value']
            except:
                pass


def parse_job_metrics(metrics_list: list):
    metrics = {}
    for job_metrics in metrics_list:
        metrics[job_metrics['name']] = job_metrics['raw_value']
    return metrics


import pathlib
from setuptools import setup, find_packages
from abm import getVersion

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="gxabm",
    version=getVersion(),
    description="Opinionated Benchmarking Automatation in Galaxy",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/galaxyproject/gxabm",
    author="Keith Suderman",
    author_email="suderman@jhu.edu",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
    ],
    packages=find_packages(),
    include_package_data=True,
    install_requires=["pathlib", "bioblend", "pyyaml", "planemo", "cloudlaunch-cli"],
    entry_points={
        "console_scripts": [
            "abm=abm.__main__:entrypoint",
        ]
    },
)


import os
import sys
import json
import yaml
import logging
import argparse
from lib import Keys, INVOCATIONS_DIR, METRICS_DIR
from lib.common import connect, Context, _get_dataset_data, _make_dataset_element, print_json
from bioblend.galaxy import GalaxyInstance, dataset_collections

log = logging.getLogger('abm')

def run_cli(context: Context, args: list):
    """
    Runs a single workflow defined by *args[0]*

    :param args: a list that contains:
    args[0] - the path to the benchmark configuration file
    args[1] - the prefix to use when creating the new history in Galaxy
    args[2] - the name of the experiment, if part of one. This is used to
              generate output folder names.

    :return: True if the workflows completed sucessfully. False otherwise.
    """
    if len(args) == 0:
        print('ERROR: no workflow configuration specified')
        return

    parser = argparse.ArgumentParser()
    parser.add_argument('workflow_path')
    parser.add_argument('-p', '--prefix')
    parser.add_argument('-e', '--experiment')
    a = parser.parse_args(args)
    #workflow_path = args[0]
    if not os.path.exists(a.workflow_path):
        print(f'ERROR: can not find workflow configuration {a.workflow_path}')
        return
    run(context, a.workflow_path, a.prefix, a.experiment)


def run(context: Context, workflow_path, history_prefix: str, experiment: str):
    # if len(args) > 1:
    #     history_prefix = args[1]
    #     if len(args) > 2:
    #         experiment = args[2].replace(' ', '_').lower()


    if os.path.exists(INVOCATIONS_DIR):
        if not os.path.isdir(INVOCATIONS_DIR):
            print('ERROR: Can not save invocation status, directory name in use.')
            sys.exit(1)
    else:
        os.mkdir(INVOCATIONS_DIR)


    if os.path.exists(METRICS_DIR):
        if not os.path.isdir(METRICS_DIR):
            print('ERROR: Can not save metrics, directory name in use.')
            return False
    else:
        os.mkdir(METRICS_DIR)

    invocations_dir = INVOCATIONS_DIR
    metrics_dir = METRICS_DIR
    if experiment is not None:
        invocations_dir = os.path.join(INVOCATIONS_DIR, experiment)
        if not os.path.exists(invocations_dir):
            os.makedirs(invocations_dir, exist_ok=True)
        metrics_dir = os.path.join(METRICS_DIR, experiment)
        if not os.path.exists(metrics_dir):
            os.makedirs(metrics_dir, exist_ok=True)

    gi = connect(context)
    workflows = parse_workflow(workflow_path)
    if not workflows:
        print(f"Unable to load any workflow definitions from {workflow_path}")
        return

    print(f"Found {len(workflows)} workflow definitions")
    for workflow in workflows:
        wf_name = workflow[Keys.WORKFLOW_ID]
        wfid = find_workflow_id(gi, wf_name)
        if wfid is None:
            print(f"Unable to load the workflow ID for {workflow[Keys.WORKFLOW_ID]}")
            return False
        else:
            print(f"Found workflow id {wfid}")
        inputs = {}
        input_names = []
        history_base_name = wfid
        if Keys.HISTORY_BASE_NAME in workflow:
            history_base_name = workflow[Keys.HISTORY_BASE_NAME]
        ref_data_size = []
        if Keys.REFERENCE_DATA in workflow:
            for spec in workflow[Keys.REFERENCE_DATA]:
                input = gi.workflows.get_workflow_inputs(wfid, spec[Keys.NAME])
                if input is None or len(input) == 0:
                    print(f'ERROR: Invalid input specification for {spec[Keys.NAME]}')
                    return False
                dsname = spec[Keys.NAME]
                #dsid = find_dataset_id(gi, dsname)
                dsdata = _get_dataset_data(gi, dsname)
                dsid = dsdata['id']
                ref_data_size.append(dsdata['size'])
                print(f"Reference input dataset {dsid}")
                inputs[input[0]] = {'id': dsid, 'src': 'hda', 'size':dsdata['size']}
                input_names.append(dsname)

        count = 0
        for run in workflow[Keys.RUNS]:
            count += 1

            # Create a new history for this run
            if Keys.HISTORY_NAME in run:
                output_history_name = f"{history_base_name} {run[Keys.HISTORY_NAME]}"
            else:
                output_history_name = f"{history_base_name} run {count}"
            new_history_name = output_history_name
            if history_prefix is not None:
                new_history_name = f"{history_prefix} {output_history_name}"
            if experiment is not None:
                new_history_name = f"{experiment} {new_history_name}"

            input_data_size = []
            if Keys.INPUTS in run and run[Keys.INPUTS] is not None:
                for spec in run[Keys.INPUTS]:
                    input = gi.workflows.get_workflow_inputs(wfid, spec[Keys.NAME])
                    if input is None or len(input) == 0:
                        print(f'ERROR: Invalid input specification for {spec[Keys.NAME]}')
                        return False

                    if 'value' in spec:
                        inputs[input[0]] = spec['value']
                        print(f"Input data value: {spec['value']}")
                    elif 'collection' in spec:
                        dsname = spec['collection']
                        input_names.append(dsname)
                        #inputs.append(dsname)
                        # dsid = find_dataset_id(gi, dsname)
                        dsdata = _get_dataset_data(gi, dsname)
                        if dsdata is None:
                            raise  Exception(f"ERROR: unable to resolve {dsname} to a dataset.")
                        dsid = dsdata['id']
                        dssize = dsdata['size']
                        input_data_size.append(dssize)
                        print(f"Input dataset ID: {dsname} [{dsid}] {dssize}")
                        inputs[input[0]] = {'id': dsid, 'src': 'hdca', 'size': dssize}
                    elif 'paired' in spec:
                        name = spec['name']
                        input_names.append(name)
                        dsdata = _get_dataset_data(gi, name)
                        if dsdata is not None:
                            print(f"Found an existing dataset named {name}")
                            print_json(dsdata)
                            # Reuse the previously defined collection
                            dsid = dsdata['id']
                            dssize = dsdata['size']
                            input_data_size.append(dssize)
                            print(f"Input dataset ID: {name} [{dsid}] {dssize}")
                            inputs[input[0]] = {'id': dsid, 'src': 'hdca', 'size': dssize}
                        else:
                            histories = gi.histories.get_histories(name=spec['history'])
                            if len(histories) == 0:
                                print(f"ERROR: History {spec['history']} not foune")
                                return
                            hid = histories[0]['id']
                            pairs = 0
                            paired_list = spec['paired']
                            for item in paired_list:
                                elements = []
                                size = 0
                                for key in item.keys():
                                    #print(f"Getting dataset for {key} = {item[key]}")
                                    value = _get_dataset_data(gi, item[key])
                                    size += value['size']
                                    elements.append(_make_dataset_element(key, value['id']))
                                description = dataset_collections.CollectionDescription(
                                    name=name,
                                    type='paired',
                                    elements=elements
                                )
                                pairs += 1
                                # print(json.dumps(description.__dict__, indent=4))
                                # pprint(description)
                                collection = gi.histories.create_dataset_collection(
                                    history_id=hid,
                                    collection_description=description
                                )
                                print(f"Input dataset paired list: {collection['id']} {size}")
                                inputs[input[0]] = {'id': collection['id'], 'src':'hdca', 'size':size}
                    elif Keys.DATASET_ID in spec:
                        dsname = spec[Keys.DATASET_ID]
                        input_names.append(dsname)
                        #inputs.append(dsname)
                        # dsid = find_dataset_id(gi, dsname)
                        dsdata = _get_dataset_data(gi, dsname)
                        if dsdata is None:
                            raise  Exception(f"ERROR: unable to resolve {dsname} to a dataset.")
                        dsid = dsdata['id']
                        dssize = dsdata['size']
                        input_data_size.append(dssize)
                        print(f"Input dataset ID: {dsname} [{dsid}] {dssize}")
                        inputs[input[0]] = {'id': dsid, 'src': 'hda', 'size': dssize}
                    else:
                        raise Exception(f'Invalid input value')
            print(f"Running workflow {wfid}")
            invocation = gi.workflows.invoke_workflow(wfid, inputs=inputs, history_name=new_history_name)
            id = invocation['id']
            invocations = gi.invocations.wait_for_invocation(id, 86400, 10, False)
            print("Waiting for jobs")
            if history_prefix is not None:
                parts = history_prefix.split()
                invocations['run'] = parts[0]
                invocations['cloud'] = parts[1]
                if len(parts) > 2:
                    invocations['job_conf'] = parts[2]
                else:
                    invocations['job_conf'] = 'Default'
            else:
                invocations['run'] = 0
                invocations['cloud'] = "N/A"
                invocations['job_conf'] = "Unknown"
            invocations['output_dir'] = metrics_dir
            invocations['inputs'] = ' '.join(input_names)
            invocations['ref_data_size'] = ref_data_size
            invocations['input_data_size'] = input_data_size
            #TODO Change this output path. (Change it to what? KS)
            output_path = os.path.join(invocations_dir, id + '.json')
            with open(output_path, 'w') as f:
                json.dump(invocations, f, indent=4)
                print(f"Wrote invocation data to {output_path}")
            wait_for_jobs(context, gi, invocations)
    print("Benchmarking run complete")
    return True



def translate(context: Context, args: list):
    if len(args) == 0:
        print('ERROR: no workflow configuration specified')
        return
    workflow_path = args[0]
    if not os.path.exists(workflow_path):
        print(f'ERROR: can not find workflow configuration {workflow_path}')
        return

    gi = connect(context)
    # wf_index,ds_index = create_rev_index(gi)
    workflows = parse_workflow(args[0])
    for workflow in workflows:
        wfid = workflow[Keys.WORKFLOW_ID]
        wfinfo = gi.workflows.show_workflow(wfid)
        if wfinfo is None or 'name' not in wfinfo:
            print(f"Warning: unable to translate workflow ID {wfid}")
        else:
            workflow[Keys.WORKFLOW_ID] = wfinfo['name']
        # if workflow[Keys.WORKFLOW_ID] in wf_index:
        #     workflow[Keys.WORKFLOW_ID] = wf_index[workflow[Keys.WORKFLOW_ID]]
        # else:
        #     print(f"Warning: no workflow id for {workflow[Keys.WORKFLOW_ID]}")
        if Keys.REFERENCE_DATA in workflow:
            for ref in workflow[Keys.REFERENCE_DATA]:
                dsid = ref[Keys.DATASET_ID]
                dataset = gi.datasets.show_dataset(dsid)
                if dataset is None:
                    print(f"Warning: could not translate dataset ID {dsid}")
                else:
                    ref[Keys.DATASET_ID] = dataset['name']
        for run in workflow[Keys.RUNS]:
            for input in run[Keys.INPUTS]:
                dsid = input[Keys.DATASET_ID]
                dataset = gi.datasets.show_dataset(dsid)
                if dataset is None:
                    print(f"Warning: could not translate dataset ID {dsid}")
                else:
                    input[Keys.DATASET_ID] = dataset['name']
    print(yaml.dump(workflows))


def validate(context: Context, args: list):
    if len(args) == 0:
        print('ERROR: no workflow configuration specified')
        return

    workflow_path = args[0]
    if not os.path.exists(workflow_path):
        print(f'ERROR: can not find workflow configuration {workflow_path}')
        return
    print(f"Validating workflow on {context.GALAXY_SERVER}")
    workflows = parse_workflow(workflow_path)
    gi = connect(context)
    total_errors = 0
    for workflow in workflows:
        wfid = workflow[Keys.WORKFLOW_ID]
        try:
            wfid = find_workflow_id(gi, wfid)
        except:
            wfid = None

        if wfid is None:
            print(f"The workflow '{workflow[Keys.WORKFLOW_ID]}' does not exist on this server.")
            return
        else:
            print(f"Workflow: {workflow[Keys.WORKFLOW_ID]} -> {wfid}")
        inputs = {}
        errors = 0
        history_base_name = wfid
        if Keys.HISTORY_BASE_NAME in workflow:
            history_base_name = workflow[Keys.HISTORY_BASE_NAME]

        if Keys.REFERENCE_DATA in workflow:
            for spec in workflow[Keys.REFERENCE_DATA]:
                input = gi.workflows.get_workflow_inputs(wfid, spec[Keys.NAME])
                if input is None or len(input) == 0:
                    print(f'ERROR: Invalid input specification for {spec[Keys.NAME]}')
                    errors += 1
                    #sys.exit(1)
                else:
                    dsid = find_dataset_id(gi, spec[Keys.DATASET_ID])
                    if dsid is None:
                        print(f"ERROR: Reference dataset not found {spec[Keys.DATASET_ID]}")
                        errors += 1
                    else:
                        print(f"Reference input dataset {spec[Keys.DATASET_ID]} -> {dsid}")
                        inputs[input[0]] = {'id': dsid, 'src': 'hda'}

        #count = 0
        for run in workflow[Keys.RUNS]:
            #count += 1
            if Keys.INPUTS in run and run[Keys.INPUTS] is not None:
                for spec in run[Keys.INPUTS]:
                    input = gi.workflows.get_workflow_inputs(wfid, spec[Keys.NAME])
                    if input is None or len(input) == 0:
                        print(f'ERROR: Invalid input specification for {spec[Keys.NAME]}')
                        errors += 1
                    else:
                        dsid = find_dataset_id(gi, spec[Keys.DATASET_ID])
                        if dsid is None:
                            print(f"ERROR: Dataset not found {spec[Keys.DATASET_ID]}")
                            errors += 1
                        else:
                            print(f"Input dataset: {spec[Keys.DATASET_ID]} -> {dsid}")
                            inputs[input[0]] = {'id': dsid, 'src': 'hda'}

        if errors == 0:
            print("This workflow configuration is valid and can be executed on this server.")
        else:
            print("---------------------------------")
            print("WARNING")
            print("The above problems need to be corrected before this workflow configuration can be used.")
            print("---------------------------------")
        total_errors += errors

    return total_errors == 0


def wait_for_jobs(context, gi: GalaxyInstance, invocations: dict):
    """ Blocks until all jobs defined in the *invocations* to complete.

    :param gi: The *GalaxyInstance** running the jobs
    :param invocations:
    :return:
    """
    wfid = invocations['workflow_id']
    hid = invocations['history_id']
    run = invocations['run']
    cloud = invocations['cloud']
    conf = invocations['job_conf']
    inputs = invocations['inputs']
    output_dir = invocations['output_dir']
    for step in invocations['steps']:
        job_id = step['job_id']
        if job_id is not None:
            retries = 3
            done = False
            while not done and retries >= 0:
                print(f"Waiting for job {job_id} on {context.GALAXY_SERVER}")
                try:
                    # TDOD Should retry if anything throws an exception.
                    status = gi.jobs.wait_for_job(job_id, 86400, 10, False)
                    data = gi.jobs.show_job(job_id, full_details=True)
                    metrics = {
                        'run': run,
                        'cloud': cloud,
                        'job_conf': conf,
                        'workflow_id': wfid,
                        'history_id': hid,
                        'inputs': inputs,
                        'metrics': data,
                        'status': status,
                        'server': context.GALAXY_SERVER,
                        'ref_data_size': invocations['ref_data_size'],
                        'input_data_size': invocations['input_data_size']
                    }
                    output_path = os.path.join(output_dir, f"{job_id}.json")
                    with open(output_path, "w") as f:
                        json.dump(metrics, f, indent=4)
                        print(f"Wrote metrics to {output_path}")
                    done = True
                except ConnectionError as e:
                    print(f"ERROR: connection dropped while waiting for {job_id}")
                    retries -= 1
                except Exception as e:
                    print(f"ERROR: {e}")
                    retries -= 1


def parse_workflow(workflow_path: str):
    if not os.path.exists(workflow_path):
        print(f'ERROR: could not find workflow file {workflow_path}')
        return None

    with open(workflow_path, 'r') as stream:
        config = yaml.safe_load(stream)
        # try:
        #     config = yaml.safe_load(stream)
        #     # print(f"Loaded {name}")
        # except yaml.YAMLError as exc:
        #     print('Error encountered parsing the YAML input file')
        #     print(exc)
        #     #TODO Don't do this...
        #     sys.exit(1)
    return config


def find_workflow_id(gi, name_or_id):
    try:
        wf = gi.workflows.show_workflow(name_or_id)
        return wf['id']
    except:
        pass

    try:
        wf = gi.workflows.get_workflows(name=name_or_id, published=True)
        return wf[0]['id']
    except:
        pass
    #print(f"Warning: unable to find workflow {name_or_id}")
    return None


def find_dataset_id(gi, name_or_id):
    # print(f"Finding dataset {name_or_id}")
    try:
        ds = gi.datasets.show_dataset(name_or_id)
        return ds['id']
    except:
        pass

    try:
        datasets = gi.datasets.get_datasets(name=name_or_id)  # , deleted=True, purged=True)
        for ds in datasets:
            if ds['state'] == 'ok' and not ds['deleted'] and ds['visible']:
                return ds['id']
            else:
                print(f"Dataset {ds['id']} is not valid.")

    except:
        print('Caught an exception')
        print(sys.exc_info())
    #print(f"Warning: unable to find dataset {name_or_id}")
    return None


from pprint import pprint
def test(context:Context, args:list):
    id = 'c90fffcf98b31cd3'
    gi = connect(context)
    inputs = gi.workflows.get_workflow_inputs(id, 'PE fastq input')
    pprint(inputs)


import json

from pprint import pprint

from bioblend.galaxy import GalaxyInstance

from common import Context, connect


def list(context: Context, args: list):
    # TODO the master API key needs to be parameterized or specified in a config file.
    context.API_KEY = "galaxypassword"
    gi = connect(context)
    user_list = gi.users.get_users()
    pprint(user_list)


def api_key(context: Context, args: list):
    print(get_api_key(context, args))


def get_api_key(context: Context, args: list):
    if len(args) == 0:
        print("ERROR: no user email given")
        return

    # TODO the master API key needs to be parameterized or specified in a config file.
    context.API_KEY = "galaxypassword"
    gi = connect(context)
    user_list = gi.users.get_users(f_email=args[0])
    if user_list is None or len(user_list) == 0:
        print("WARNING: no such user")
        return
    if len(user_list) > 1:
        print("WARNING: more than one user with that email address")
        return
    id = user_list[0]['id']
    key = gi.users.get_user_apikey(id)
    if key is None or key == 'Not available.':
        # print(f"Creating API key for {args[0]}")
        key = gi.users.create_user_apikey(id)
    return key


def create(context: Context, args:list):
    if len(args) != 3:
        print("ERROR: Please specify the username, email, and password")
        return

    name = args[0]
    email: str = args[1]
    password = args[2]
    if not '@' in email:
        print(f"ERROR: {email} does not look like a valid email address")
        return

    # TODO the master API key needs to be parameterized or specified in a config file.
    context.API_KEY = "galaxypassword"
    gi = connect(context)
    user_record = gi.users.create_local_user(name, email, password)
    id = user_record['id']
    key = gi.users.create_user_apikey(id)
    result = {
        'name': name,
        'email': email,
        'id': id,
        'key': key
    }
    print(json.dumps(result, indent=4))


def show(context: Context, args:list):
    if len(args) == 0:
        print("ERROR: no user email given")
        return

    # TODO the master API key needs to be parameterized or specified in a config file.
    context.API_KEY = "galaxypassword"
    gi = connect(context)
    id = _get_user_id(gi, args[0])
    if id is None:
        return
    result = gi.users.show_user(id)
    print(json.dumps(result, indent=4))


def usage(context:Context, args:list):
    if len(args) == 0:
        print("ERROR: no user email given")
        return

    # TODO the master API key needs to be parameterized or specified in a config file.
    context.API_KEY = "galaxypassword"
    gi = connect(context)
    id = _get_user_id(gi, args[0])
    if id is None:
        return
    user_data = gi.users.show_user(id)
    usage = user_data['total_disk_usage']
    MB = 1048576
    GB = 1073741824
    if usage < MB:
        num = usage / 1024
        unit = 'K'
    elif usage < GB:
        num = usage / MB
        unit = 'M'
    else:
        num = usage / GB
        unit = 'G'
    print(f"{num:.2f}{unit}")


def _get_user_id(gi: GalaxyInstance, name_or_email: str) -> str:
    user_list = gi.users.get_users(f_email=name_or_email)
    if user_list is None or len(user_list) == 0:
        print("WARNING: no such user")
        return None
    if len(user_list) > 1:
        print("WARNING: more than one user with that email address. Returning the first")
    return user_list[0]['id']
